const fs = require('fs');
const path = require('path');
const { WLD } = require('./WLDUtils');
const sqlite3 = require('sqlite3').verbose();

const dbPath = '/home/pi/wld/database/database.db';

function readQuery(queryPath) {
    const filePath = path.join(__dirname, '../DEV/queries', queryPath);
    return fs.readFileSync(filePath, 'utf-8');
}

function handleError(err, text) {
    if (err) {
        console.error(text, err.message);
    }
}

const sqlFiles = {
    csv: 'get/csv.sql',
    pdf: 'get/pdf.sql',
    print: 'get/print.sql',
    essay: 'get/essay.sql',
    users: 'get/users.sql',
    qrcode: 'get/qrcode.sql',
    essays: 'get/essays.sql',
    equips: 'get/equips.sql',
    default: 'get/default.sql',
    create: {
        users: 'create/users.sql',
        equips: 'create/equips.sql',
        prints: 'create/prints.sql',
        essays: 'create/essays.sql'
    }
};

function runQuery(db, query, params = []) {
    return new Promise((resolve, reject) => {
        db.run(query, params, (err) => {
            if (err) {
                handleError(err, 'Erro ao inserir dados:');
                reject(err);
            } else {
                resolve(true);
            }
        });
    });
}

class WLDBD {

    static database_location = '/home/pi/wld/database/database.db';
    static clean_database_location = '/home/pi/wld/database/database_clean.db';

    static async accessDatabase(callback) {
        const db = new sqlite3.Database(dbPath, sqlite3.OPEN_READWRITE, (err) => {
            if (err) {
                console.error('Erro ao abrir o banco de dados:', err.message);
                return;
            }
        });

        try {
            await callback(db);
        } finally {
            db.close((err) => handleError(err, 'Erro ao fechar o banco de dados:'));
        }
    }

    static async createDatabase() {
        await this.accessDatabase(async (db) => {
            const createQueries = [
                sqlFiles.create.users,
                sqlFiles.create.equips,
                sqlFiles.create.prints,
                sqlFiles.create.essays
            ];

            for (const query of createQueries) {
                db.run(readQuery(query), (err) => handleError(err, 'Erro ao criar tabela:'));
            }
        });
    }

    static async insertNewData(table, data) {
        const insertQueries = {
            prints: {
                query: `INSERT INTO prints (id_essay) VALUES (?)`,
                columns: ['id_essay']
            },
            users: {
                query: `INSERT INTO users (name, pwd, type) VALUES (?, ?, ?)`,
                columns: ['name', 'pwd', 'type']
            },
            equips: {
                query: `INSERT INTO equips (brand, model, code) VALUES (?, ?, ?)`,
                columns: ['brand', 'model', 'code']
            },
            essays: {
                query: `INSERT INTO essays (result, type, batch, id_user, id_equip, cavity, duration, date, cycle, ampoule) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)`,
                columns: ['result', 'type', 'batch', 'id_user', 'id_equip', 'cavity', 'duration', 'date', 'cycle', 'ampoule']
            }
        };
    
        const { query, columns } = insertQueries[table];
    
        const values = columns.map(column => data[column]);
    
        await this.accessDatabase(async (db) => {
            await runQuery(db, query, values);
        });
    }

    static async editData(table, { _old, _new, _id, _type }) {
        const updateQuery = `UPDATE ${table} SET ${_old} = '${_new}' WHERE ${_type} = ?`;

        console.log('SQL: ' + updateQuery);
        
        await this.accessDatabase(async (db) => {
            db.run(updateQuery, [_id], (err) => handleError(err, 'Erro ao atualizar dados:'));
        });
    }

    static async correctDatabase() {
        const updateQuery = `UPDATE essays SET result = 3 WHERE duration = 0 AND result = 0`;

        await this.accessDatabase(async (db) => {
            db.run(updateQuery, (err) => handleError(err, 'Erro ao corrigir dados:'));
        });
    }

    static async deleteQueue(id) {
        const deleteQuery = `DELETE FROM prints WHERE id_essay = ?`;

        await this.accessDatabase(async (db) => {
            db.run(deleteQuery, [id], (err) => handleError(err, 'Erro ao deletar dados da fila:'));
        });
    }

    static async deleteData(table, id, selector = 'id_essay') {
        const deleteQuery = `DELETE FROM ${table} WHERE ${selector} = ?`;

        await this.accessDatabase(async (db) => {
            db.run(deleteQuery, [id], (err) => handleError(err, 'Erro ao deletar dados:'));
        });
    }
    
    static async getData(table, id, type) {
        const getQueries = {
            default:    readQuery(sqlFiles.default),
            qrcode:     readQuery(sqlFiles.qrcode),
            prints:     readQuery(sqlFiles.print),
            essays:     readQuery(sqlFiles.essay),

            verify: `SELECT count(${type}) as "count" FROM essays WHERE ${type} = ?`,
        };

        return new Promise((resolve, reject) => {
            this.accessDatabase(async (db) => {
                db.get(getQueries[table], [id], (err, rows) => {
                    if (err) {
                        console.error('Erro ao obter dados:', err.message);
                        reject(err);
                    } else {
                        resolve(rows);
                    }
                });
            });
        });
    }

    static async getDatas(table, id) {
        const getQueries = {
            users:      readQuery(sqlFiles.users),
            equips:     readQuery(sqlFiles.equips),
            all_pdfs:   readQuery(sqlFiles.pdf),
            csv_essays: readQuery(sqlFiles.csv),
            all_essays: readQuery(sqlFiles.essays),
        };

        return new Promise((resolve, reject) => {
            this.accessDatabase(async (db) => {
                db.all(getQueries[table], [id], (err, rows) => {
                    if (err) {
                        console.error('Erro ao obter dados:', err.message);
                        reject(err);
                    } else {
                        resolve(rows);
                    }
                });
            });
        });
    }
}

module.exports = WLDBD;