const { WLD } = require('./WLDUtils');
const WLDConfig = require('./WLDConfig');
const WLDLanguage = require('./WLDLanguage');
const { types: Types, printer: ThermalPrinter } = require("node-thermal-printer");

const options = {
    width: 32, 
    type: Types.EPSON,
    interface: '/dev/usb/lp0',
    options: { timeout: 5000 },
};

const PICS_FOLDER   = '/home/pi/wld/pics/';
const COMMON_FOLDER = PICS_FOLDER + 'common/';
const PRINT_TEST    = COMMON_FOLDER + 'print_test_';
const NOT_FOUND     = COMMON_FOLDER + 'not_found_';
const CANCELED      = COMMON_FOLDER + 'canceled_';

const ASSETS_FOLDER = '/home/pi/wld/lummina/src/assets/';
const LUMMINA_LOGO  = ASSETS_FOLDER + 'print/lumina.png';

class WLDPrinter {
    constructor () {
        this.isPrinting = false;
        this.isConnected = false;
        this.printer = new ThermalPrinter(options);
        this.images = {
            print: {
                "PORTUGUÊS": PRINT_TEST + 'portuguese.png',
                "ESPANHOL":  PRINT_TEST + 'spanish.png',
                "INGLÊS":    PRINT_TEST + 'english.png'
            },
            not: {
                "PORTUGUÊS": NOT_FOUND + 'portuguese.png',
                "ESPANHOL":  NOT_FOUND + 'spanish.png',
                "INGLÊS":    NOT_FOUND + 'english.png'
            },
            canceled: {
                "PORTUGUÊS": CANCELED + 'portuguese.png',
                "ESPANHOL":  CANCELED + 'spanish.png',
                "INGLÊS":    CANCELED + 'english.png'
            }
        };

        this.mode = {
            'medcontrol':   ASSETS_FOLDER + 'html/logo_med_2.png',
            'tecil':        ASSETS_FOLDER + 'html/logo_tecil_2.png',
            'maxximed':     ASSETS_FOLDER + 'html/logo_maxximed_2.png'
        } 

        this.type = WLDConfig.read('type', 'mode', 'maxximed');
    }

    translate = (id) => WLDLanguage.translate_printer(id);

    readPrint = () => this.isPrinting;
    setPrint  = () => this.isPrinting = true;
    freePrint = () => this.isPrinting = false;

    generateOptions(translate, param) {
        return { align: 'LEFT', text: `${this.translate(translate)}: ${param}`, bold: true }
    }

    async isConected() {
        try { 
            return await this.printer.isPrinterConnected();
        } catch (e) { 
            WLD.log('PRINT',`error, not conected.`);
        }
    }

    async testPrinter({ number = 10, date, hour }) {

        this.printer.alignLeft();

        try { 
            await this.printer.printImage(LUMMINA_LOGO);
        } catch (e) { 
            return WLD.log('PRINT',`error while printing: ${e}`);
        }

        this.printer.newLine();
        this.printer.alignCenter();
        this.printer.setTypeFontB();
        this.printer.println(`         ${this.translate(115)} ${number}`);
        this.printer.setTypeFontA();
        this.printer.alignLeft();

        try { 
            await this.printer.printImage(this.images['print'][WLDLanguage.getCurrentPrinter()]);
        } catch (e) { 
            await this.printer.printImage(this.images['print'][WLDLanguage.getCurrentPrinter()]);
        }

        this.printer.newLine();
        
        const { 
            first_column: date_cols, 
            second_column: date_cols_2, 
            third_column: hour_cols, 
            fourth_column: hour_cols_2  
        } = this.calculateTwoColumnLength(this.translate(35), date, this.translate(37));

        this.printer.tableCustom([
            { align: 'LEFT',  cols: date_cols,   text: `${this.translate(35)}: `, bold: true },
            { align: 'LEFT',  cols: date_cols_2, text: `${date}` },
            { align: 'RIGHT', cols: hour_cols,   text: `${this.translate(37)}: `, bold: true },
            { cols: hour_cols_2, text: `${hour}` }
        ]);
        
        this.printer.alignLeft();

        this.type = WLDConfig.read('type', 'mode', 'maxximed');

        let image_print = (this.mode[this.type] == undefined) 
            ? this.mode['maxximed'] 
            : this.mode[this.type];

        try { 
            await this.printer.printImage(image_print);
        } catch (e) { 
            return WLD.log('PRINT',`error while printing: ${e}`);
        }

        this.printer.newLine();
        this.printer.newLine();
        this.printer.newLine();
        this.printer.newLine();

        try { 
            await this.printer.execute();
        } catch (e) {
            this.isPrinting = false;
            this.printer.clear();
            
            return WLD.log('PRINT',`error while printing: ${e}`);
        }
        
        this.printer.clear();
        this.isPrinting = false;

        return WLD.log('PRINT',`Printer Test, has Been Printed Successfully`);
    }

    calculateOneColumnLength(string) {
        return { first_column: string.length + 2, second_column: (options.width - 2) - string.length }
    }

    calculateTwoColumnLength(first_string, second_string, third_string) {

        const space_used = (first_string.length + 2) + (third_string.length + 2);
        const space_left = options.width - space_used;
        const space_four = space_left - (second_string.length + 1);

        return { 
            first_column:  first_string.length + 2, 
            second_column: second_string.length + 1,
            third_column:  third_string.length + 2, 
            fourth_column: space_four
        }
    }

    async print({ number, type, ampoule, cycle, batch, equip, user, time, result, date, hour, image, cavity, incub }, resQR) {
        this.printer.alignLeft();
        
        try { 
            await this.printer.printImage(LUMMINA_LOGO);
        } catch (e) { 
            return WLD.log('PRINT',`error while printing: ${e}`);
        }

        this.printer.newLine();
        this.printer.alignCenter();
        this.printer.setTypeFontB();
        this.printer.println(`         ${this.translate(115)} ${number}`);
        this.printer.setTypeFontA();

        this.printer.newLine();
        this.printer.tableCustom([this.generateOptions(106, type)]);
        this.printer.tableCustom([this.generateOptions(107, ampoule)]);
        this.printer.tableCustom([this.generateOptions(125, cycle)]);
        this.printer.tableCustom([this.generateOptions(108, batch)]);
        this.printer.tableCustom([this.generateOptions(126, cavity)]);
        this.printer.tableCustom([this.generateOptions(109, equip)]);
        this.printer.tableCustom([this.generateOptions(110, user)]);
        this.printer.tableCustom([this.generateOptions(111, time)]);
        this.printer.tableCustom([this.generateOptions(112, result)]);
        this.printer.tableCustom([this.generateOptions(113, incub.toString().padStart(6,0))]);
        
        this.printer.newLine();
        this.printer.alignLeft();

        try { 
            (result == this.translate(52))
                ? await this.printer.printImage(this.images["canceled"][WLDLanguage.getCurrentPrinter()])
                : await this.printer.printImage(PICS_FOLDER + `print/${image}.png`);

        } catch (e) { 
            await this.printer.printImage(this.images["not"][WLDLanguage.getCurrentPrinter()]);
        }
        
        this.printer.newLine();
        this.printer.setTypeFontB();
        this.printer.alignLeft();
        this.printer.println(`${this.translate(114)}`);
        this.printer.setTypeFontA();

        if(WLDConfig.read('settings', 'print_qrcode', false) && resQR) {
            try { 
                await this.printer.printImage(PICS_FOLDER + 'qrcode.png');
            } catch (e) { 
                return WLD.log('PRINTER', `error while printig qrcode: ${e}`);
            }
        }

        this.printer.newLine();

        const { 
            first_column: date_cols, 
            second_column: date_cols_2, 
            third_column: hour_cols, 
            fourth_column: hour_cols_2  
        } = this.calculateTwoColumnLength(this.translate(35), date, this.translate(37));

        this.printer.tableCustom([
            { align: 'LEFT',  cols: date_cols,   text: `${this.translate(35)}: `, bold: true },
            { align: 'LEFT',  cols: date_cols_2, text: `${date}` },
            { align: 'RIGHT', cols: hour_cols,   text: `${this.translate(37)}: `, bold: true },
            { cols: hour_cols_2, text: `${hour}` }
        ]);
        
        this.printer.alignLeft();

        this.type = WLDConfig.read('type', 'mode', 'maxximed');

        let image_print = (this.mode[this.type] == undefined) 
            ? this.mode['maxximed'] 
            : this.mode[this.type];

        try { 
            await this.printer.printImage(image_print);
        } catch (e) { 
            return WLD.log('PRINT',`error while printing: ${e}`);
        }
        
        this.printer.newLine();
        this.printer.newLine();
        this.printer.newLine();
        this.printer.newLine();

        try { 
            await this.printer.execute();
        } catch (e) {
            this.isPrinting = false;
            this.printer.clear();

            return WLD.log('PRINT',`error while printing: ${e}`);
        }
        
        this.printer.clear();
        this.isPrinting = false;

        return WLD.log('PRINT',`Registry: ${incub.toString().padStart(6,0)} has Been Printed Successfully`);
    }
}

module.exports = WLDPrinter;